// Copyright © Microsoft Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.register.provider.azure.subscriber.pubsub;

import java.util.concurrent.atomic.AtomicReference;

import org.opengroup.osdu.azure.eventgridmanager.EventGridManagerFactory;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.register.provider.azure.di.AzureBootstrapConfig;
import org.opengroup.osdu.register.provider.azure.util.PartitionHelperEventGrid;
import org.opengroup.osdu.register.utils.AppServiceConfig;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.stereotype.Component;

import com.microsoft.azure.CloudException;
import com.microsoft.azure.arm.model.Indexable;
import com.microsoft.azure.management.eventgrid.v2020_04_01_preview.WebHookEventSubscriptionDestination;

@Component
@ConditionalOnProperty(value = "azure.eventGrid.enabled", havingValue = "true", matchIfMissing = false)
public class PushSubscription implements IPubsubSubscription {

    private static final String RESOURCE_PROVISIONING_ERROR_MESSAGE = "Resource cannot be updated during provisioning";

    @Autowired
    private EventGridManagerFactory factory;
    @Autowired
    private DpsHeaders headers;
    @Autowired
    private AzureBootstrapConfig azureBootstrapConfig;
    @Autowired
    private JaxRsDpsLog logger;
    @Autowired
    private AppServiceConfig serviceConfig;
    @Autowired
    private PartitionHelperEventGrid partitionHelper;

    public void createSubscription(String subscriptionId, String topicAliasName) {
        String topicName = partitionHelper.getFullyQualifiedTopicName(topicAliasName);
        WebHookEventSubscriptionDestination subscriptionDestination = new WebHookEventSubscriptionDestination();
        subscriptionDestination.withEndpointUrl(serviceConfig.getRecordsChangePubsubEndpoint());
        subscriptionDestination.withAzureActiveDirectoryTenantId(azureBootstrapConfig.getAzureTenantId());
        subscriptionDestination.withAzureActiveDirectoryApplicationIdOrUri(azureBootstrapConfig.getAzureAppResourceId());

        String scope = partitionHelper.retrieveScope((topicName));
        AtomicReference<Throwable> error = new AtomicReference<>();

        factory.getManager(headers.getPartitionId()).eventSubscriptions().define(subscriptionId)
                .withScope(scope)
                .withDestination(subscriptionDestination)
                .createAsync()
                .toBlocking()
                .subscribe(
                        (Indexable indexable) -> {},
                        error::set,
                        () -> logger.debug(String.format("Push Subscription with id %s created successfully", subscriptionId))
                );

        if(error.get() != null) {
            if(error.get().getMessage().equals(RESOURCE_PROVISIONING_ERROR_MESSAGE)) {
                logger.error("Another request is trying to create the same Push subscription");
                throw new AppException(409, "Conflict", "Another request is trying to create the same Push subscription");
            }
            else {
                logger.error("Creating Push Subscription failed with error: " + error.get().toString());
                throw new AppException(500, "Server Error", "Unexpected error creating Push subscription");
            }
        }

    }

    public void deleteSubscription(String subscriptionId, String topicAliasName) {
        String topicName = partitionHelper.getFullyQualifiedTopicName(topicAliasName);
        String scope = partitionHelper.retrieveScope(topicName);
        AtomicReference<Throwable> error = new AtomicReference<>();
        factory.getManager(headers.getPartitionId()).eventSubscriptions().deleteAsync(scope, subscriptionId)
                .subscribe(() -> logger.debug(String.format("Push Subscription with id %s deleted successfully", subscriptionId)), error::set);

        if(error.get() != null) {
            if(error.get() instanceof CloudException) {
                CloudException cloudException = (CloudException) error.get();
                logger.error(cloudException.toString());
                throw new AppException(cloudException.response().code(), cloudException.body().code(), cloudException.body().message());
            }
            else {
                logger.error("Deleting Push Subscription failed with error: " + error.get().toString());
                throw new AppException(500, "Server Error", "Unexpected error deleting Push subscription");
            }
        }
    }

    public void checkIfSubscriptionExists(String subscriptionId, String topicAliasName) {
        String topicName = partitionHelper.getFullyQualifiedTopicName(topicAliasName);
        String scope = partitionHelper.retrieveScope(topicName);
        AtomicReference<Throwable> error = new AtomicReference<>();
        factory.getManager(headers.getPartitionId()).eventSubscriptions().getAsync(scope, subscriptionId)
                .subscribe(
                        (Indexable indexable) -> {},
                        error::set,
                        () -> logger.debug(String.format("Push Subscription with id %s fetched successfully", subscriptionId))
                );

        if(error.get() != null) {
            if(error.get() instanceof NullPointerException) {
                logger.error(String.format("Push Subscription with id %s does not exist.", subscriptionId));
                throw new AppException(404, "Not found", String.format("Push Subscription with id %s does not exist.", subscriptionId));
            }
            else {
                logger.error("Fetching Push subscription failed with error: " + error.get().toString());
                throw new AppException(500, "Server Error", "Unexpected error while Fetching Push subscription");
            }
        }
    }
}
