// Copyright © Microsoft Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.register.provider.azure.ddms;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import org.opengroup.osdu.azure.cosmosdb.CosmosStore;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.register.ddms.model.Ddms;
import org.opengroup.osdu.register.ddms.model.RegisteredInterface;
import org.opengroup.osdu.register.provider.azure.di.AzureBootstrapConfig;
import org.opengroup.osdu.register.provider.azure.di.CosmosContainerConfig;

import java.util.Collections;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.doThrow;
import static org.mockito.Mockito.lenient;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;

@ExtendWith(MockitoExtension.class)
public class DdmsRepositoryTest {

    private static final String dataPartitionId = "data-partition-id";
    private static final String ddmsId = "id-1";
    private static final String ddmsContainer = "ddms-container";
    private static final String cosmosDatabase = "cosmos-database";

    @Mock
    private DpsHeaders dpsHeaders;

    @Mock
    private CosmosStore cosmosStore;

    @Mock
    private AzureBootstrapConfig azureBootstrapConfig;

    @Mock
    private CosmosContainerConfig cosmosContainerConfig;

    @Mock
    private JaxRsDpsLog logger;

    @InjectMocks
    private DdmsRepository repo;


    @BeforeEach
    void init() {
        lenient().doReturn(dataPartitionId).when(dpsHeaders).getPartitionId();
        lenient().when(azureBootstrapConfig.getCosmosDBName()).thenReturn(cosmosDatabase);
        lenient().when(cosmosContainerConfig.getDdmsContainerName()).thenReturn(ddmsContainer);
    }

    @Test
    public void createDdmsThrows409WhenIdExists() {
        Ddms ddms = createDdms();
        doThrow(new AppException(409, "Reason", "Message")).when(cosmosStore).createItem(eq(dataPartitionId), eq(cosmosDatabase), eq(ddmsContainer), eq(dataPartitionId), any());

        AppException exception = assertThrows(AppException.class, () -> {
            repo.create(ddms);
        });

        assertEquals(409, exception.getError().getCode());
        assertEquals("Conflict", exception.getError().getReason());
        assertEquals(String.format("A DDMS already exists with the same id: %s", ddms.getId()), exception.getError().getMessage());
        verify(cosmosContainerConfig, times(1)).getDdmsContainerName();
        verify(dpsHeaders, times(3)).getPartitionId();
        verify(logger, times(1)).error(String.format("A DDMS already exists with the same id: %s", ddms.getId()));
    }

    @Test
    public void createDdmsDocumentInsertedSuccessfully() {
        Ddms ddms = createDdms();
        doNothing().when(cosmosStore).createItem(eq(dataPartitionId), eq(cosmosDatabase), eq(ddmsContainer), eq(dataPartitionId), any());

        Ddms output = repo.create(ddms);

        assertEquals(ddms, output);
        verify(azureBootstrapConfig, times(1)).getCosmosDBName();
        verify(cosmosContainerConfig, times(1)).getDdmsContainerName();
        verify(dpsHeaders, times(3)).getPartitionId();
    }

    @Test
    public void getDdmsThrows404WhenIdNotFound() {

        doReturn(Optional.empty()).when(cosmosStore).findItem(eq(dataPartitionId), eq(cosmosDatabase), eq(ddmsContainer), eq(ddmsId), eq(dataPartitionId), any());

        AppException exception = assertThrows(AppException.class, () -> {
            repo.get(ddmsId);
        });

        assertEquals(404, exception.getError().getCode());
        assertEquals("Not found", exception.getError().getReason());
        assertEquals("DDMS with id id-1 does not exist.", exception.getError().getMessage());
        verify(cosmosStore, times(1)).findItem(dataPartitionId, cosmosDatabase, ddmsContainer, ddmsId, dataPartitionId, DdmsDoc.class);
        verify(cosmosContainerConfig, times(1)).getDdmsContainerName();
        verify(dpsHeaders, times(2)).getPartitionId();
        verify(logger, times(1)).error("DDMS with id id-1 does not exist.");
    }

    @Test
    public void getDdmsItemRetrievedSuccessfully() {

        Ddms ddms = createDdms();
        DdmsDoc doc = new DdmsDoc(ddms, dataPartitionId);

        doReturn(Optional.of(doc)).when(cosmosStore).findItem(eq(dataPartitionId), eq(cosmosDatabase), eq(ddmsContainer), eq(ddmsId), eq(dataPartitionId), any());

        Ddms output = repo.get(ddmsId);
        ddms.setCreatedDateTimeEpoch(output.getCreatedDateTimeEpoch());

        assertEquals(ddms, output);
        verify(cosmosStore, times(1)).findItem(dataPartitionId, cosmosDatabase, ddmsContainer, ddmsId, dataPartitionId, DdmsDoc.class);
        verify(azureBootstrapConfig, times(1)).getCosmosDBName();
        verify(cosmosContainerConfig, times(1)).getDdmsContainerName();
        verify(dpsHeaders, times(2)).getPartitionId();
    }

    @Test
    public void deleteDdmsReturnsFalseWhenIdNotFound() {

        doThrow(new AppException(404, "Reason", "Message")).when(cosmosStore).deleteItem(eq(dataPartitionId), eq(cosmosDatabase), eq(ddmsContainer), eq(ddmsId), eq(dataPartitionId));

        boolean output = repo.delete(ddmsId);

        assertFalse(output);
        verify(cosmosStore, times(1)).deleteItem(dataPartitionId, cosmosDatabase, ddmsContainer, ddmsId, dataPartitionId);
        verify(azureBootstrapConfig, times(1)).getCosmosDBName();
        verify(cosmosContainerConfig, times(1)).getDdmsContainerName();
        verify(dpsHeaders, times(2)).getPartitionId();
        verify(logger, times(1)).error(String.format("DDMS with id %s does not exist.", ddmsId));
    }

    @Test
    public void deleteDdmsReturnsTrueWhenDeletedSuccessfully() {

        doNothing().when(cosmosStore).deleteItem(eq(dataPartitionId), eq(cosmosDatabase), eq(ddmsContainer), eq(ddmsId), eq(dataPartitionId));

        boolean output = repo.delete(ddmsId);

        assertTrue(output);
        verify(cosmosStore, times(1)).deleteItem(dataPartitionId, cosmosDatabase, ddmsContainer, ddmsId, dataPartitionId);
        verify(azureBootstrapConfig, times(1)).getCosmosDBName();
        verify(cosmosContainerConfig, times(1)).getDdmsContainerName();
        verify(dpsHeaders, times(2)).getPartitionId();
    }

    @Test
    public void queryDdmsReturnItemsSuccessfully() {

        Ddms ddms1 = createDdms();
        Ddms ddms2 = createDdms();
        ddms2.setId("id-2");

        DdmsDoc doc1 = new DdmsDoc(ddms1, dataPartitionId);
        DdmsDoc doc2 = new DdmsDoc(ddms2, dataPartitionId);

        List<Ddms> ddmsList = Stream.of(ddms1, ddms2).collect(Collectors.toList());
        List<DdmsDoc> ddmsDocList = Stream.of(doc1, doc2).collect(Collectors.toList());

        doReturn(ddmsDocList).when(cosmosStore).queryItems(eq(dataPartitionId), eq(cosmosDatabase), eq(ddmsContainer), any(), any(), any());

        List<Ddms> output = repo.query("type");

        ddmsList.get(0).setCreatedDateTimeEpoch(output.get(0).getCreatedDateTimeEpoch());
        ddmsList.get(1).setCreatedDateTimeEpoch(output.get(1).getCreatedDateTimeEpoch());

        assertEquals(ddmsList, output);
        verify(azureBootstrapConfig, times(1)).getCosmosDBName();
        verify(cosmosContainerConfig, times(1)).getDdmsContainerName();
        verify(dpsHeaders, times(2)).getPartitionId();
    }

    private Ddms createDdms() {
        Ddms ddms = new Ddms();
        RegisteredInterface ri = new RegisteredInterface();
        ri.setEntityType("type");
        ri.setSchema(Collections.singletonMap("first", "second"));

        ddms.setId(ddmsId);
        ddms.setName("mock-name");
        ddms.setDescription("mock-description");
        ddms.setContactEmail("mock-contactEmail");
        ddms.getInterfaces().add(ri);
        return ddms;
    }

}
