// Copyright © Microsoft Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.register.provider.azure.subscriber;

import com.microsoft.azure.CloudError;
import com.microsoft.azure.CloudException;
import com.microsoft.azure.arm.model.Indexable;
import com.microsoft.azure.management.eventgrid.v2020_04_01_preview.EventSubscription;
import com.microsoft.azure.management.eventgrid.v2020_04_01_preview.EventSubscriptions;
import com.microsoft.azure.management.eventgrid.v2020_04_01_preview.implementation.EventGridManager;

import okhttp3.ResponseBody;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import org.opengroup.osdu.azure.eventgridmanager.EventGridManagerFactory;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.register.provider.azure.di.AzureBootstrapConfig;
import org.opengroup.osdu.register.provider.azure.subscriber.pubsub.PushSubscription;
import org.opengroup.osdu.register.provider.azure.util.PartitionHelperEventGrid;
import org.opengroup.osdu.register.utils.AppServiceConfig;

import retrofit2.Response;
import rx.Completable;
import rx.Observable;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.lenient;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

@ExtendWith(MockitoExtension.class)
public class PushSubscriptionTest {

    private static final String pushEndpoint = "http://testing-env.com/api/v1/records-changed";
    private static final String subscriptionId = "subscription-id";
    private static final String resourceGroupName = "resource-group";
    private static final String topicName = "topic-name";
    private static final String errorMessage = "some error message";
    private static final String tenantId = "tenant-id";
    private static final String appResourceId = "app-resource-id";
    private static final String scope = String.format("/subscriptions/%s/resourceGroups/%s/providers/Microsoft.EventGrid/topics/%s", subscriptionId, resourceGroupName, topicName);
    public static final String PARTITION_ID = "partition-id";
    private final Observable<Indexable> observable = Observable.from(new Indexable[]{});
    private final Observable<Indexable> observableError = Observable.error(new Throwable(errorMessage));
    private final Observable<EventSubscription> observableGetSubscription = Observable.from(new EventSubscription[]{});
    private final Observable<EventSubscription> observableErrorNullPointerExceptionGetSubscription = Observable.error(new NullPointerException());
    private final Observable<EventSubscription> observableError500ExceptionGetSubscription = Observable.error(new Throwable(errorMessage));
    private final Completable completable = Completable.complete();

    @Mock
    private EventGridManager eventGridManager;
    @Mock
    private EventGridManagerFactory factory;
    @Mock
    private DpsHeaders headers;
    @Mock
    private AzureBootstrapConfig azureBootstrapConfig;

    @Mock
    private JaxRsDpsLog logger;

    @Mock
    private EventSubscriptions eventSubscriptions;

    @Mock
    private EventSubscription.DefinitionStages.Blank eventSubscription;

    @Mock
    private EventSubscription.DefinitionStages.WithCreate eventSubscriptionWithCreate;

    @Mock
    private AppServiceConfig serviceConfig;

    @Mock
    private Response<ResponseBody> response;

    @Mock
    private CloudError cloudError;

    @Mock
    private PartitionHelperEventGrid partitionHelper;

    @InjectMocks
    private PushSubscription pushSubscription;

    @BeforeEach
    public void init() {
        lenient().when(partitionHelper.retrieveResourceGroup()).thenReturn(resourceGroupName);
        String format = String.format("/subscriptions/%s/resourceGroups/%s/providers/Microsoft.EventGrid/topics/%s",
                subscriptionId, resourceGroupName, topicName);
        lenient().when(partitionHelper.retrieveScope(topicName)).thenReturn(format);
        lenient().when(azureBootstrapConfig.getAzureTenantId()).thenReturn(tenantId);
        lenient().when(azureBootstrapConfig.getAzureAppResourceId()).thenReturn(appResourceId);
        lenient().when(serviceConfig.getRecordsChangePubsubEndpoint()).thenReturn(pushEndpoint);
        lenient().when(headers.getPartitionId()).thenReturn(PARTITION_ID);
        lenient().when(factory.getManager(PARTITION_ID)).thenReturn(eventGridManager);
        lenient().when(eventGridManager.eventSubscriptions()).thenReturn(eventSubscriptions);
        lenient().when(eventSubscriptions.define(subscriptionId)).thenReturn(eventSubscription);
        lenient().when(eventSubscription.withScope(scope)).thenReturn(eventSubscriptionWithCreate);
        lenient().when(eventSubscriptionWithCreate.withDestination(any())).thenReturn(eventSubscriptionWithCreate);
        lenient().when(partitionHelper.getFullyQualifiedTopicName(topicName)).thenReturn(topicName);

    }

    @Test
    public void shouldCreatePushSubscription() {
        when(eventSubscriptionWithCreate.createAsync()).thenReturn(observable);
        pushSubscription.createSubscription(subscriptionId, topicName);

        verify(serviceConfig, times(1)).getRecordsChangePubsubEndpoint();
        verify(eventGridManager, times(1)).eventSubscriptions();
        verify(eventSubscriptions, times(1)).define(subscriptionId);
        verify(eventSubscription, times(1)).withScope(scope);
        verify(logger, times(1)).debug(String.format("Push Subscription with id %s created successfully", subscriptionId));
    }

    @Test
    public void shouldReturn500WhenPushSubscriptionCreationFailed() {

        when(eventSubscriptionWithCreate.createAsync()).thenReturn(observableError);

        AppException exception = assertThrows(AppException.class, () -> {
            pushSubscription.createSubscription(subscriptionId, topicName);
        });

        assertNotNull(exception);
        assertEquals(500, exception.getError().getCode());
        verify(serviceConfig, times(1)).getRecordsChangePubsubEndpoint();
        verify(eventGridManager, times(1)).eventSubscriptions();
        verify(eventSubscriptions, times(1)).define(subscriptionId);
        verify(eventSubscription, times(1)).withScope(scope);
    }

    @Test
    public void shouldDeletePushSubscription() {

        when(eventSubscriptions.deleteAsync(scope, subscriptionId)).thenReturn(completable);

        pushSubscription.deleteSubscription(subscriptionId, topicName);

        verify(eventGridManager, times(1)).eventSubscriptions();
        verify(logger, times(1)).debug(String.format("Push Subscription with id %s deleted successfully", subscriptionId));
    }

    @Test
    public void shouldThrow409ExceptionWhenDeletingPushSubscription() {

        when(response.code()).thenReturn(409);
        when(cloudError.code()).thenReturn(errorMessage);
        when(cloudError.message()).thenReturn(errorMessage);
        Completable completableError = Completable.error(new CloudException(errorMessage, response, cloudError));

        when(eventSubscriptions.deleteAsync(scope, subscriptionId)).thenReturn(completableError);

        AppException exception = assertThrows(AppException.class, () -> {
            pushSubscription.deleteSubscription(subscriptionId, topicName);
        });

        assertNotNull(exception);
        assertEquals(409, exception.getError().getCode());
        assertEquals(errorMessage, exception.getError().getReason());
        assertEquals(errorMessage, exception.getError().getMessage());
        verify(eventGridManager, times(1)).eventSubscriptions();
    }

    @Test
    public void shouldThrow500ExceptionWhenDeletingPushSubscription() {

        Completable completableError = Completable.error(new Throwable(errorMessage));
        when(eventSubscriptions.deleteAsync(scope, subscriptionId)).thenReturn(completableError);

        AppException exception = assertThrows(AppException.class, () -> {
            pushSubscription.deleteSubscription(subscriptionId, topicName);
        });

        assertNotNull(exception);
        assertEquals(500, exception.getError().getCode());
        assertEquals(500, exception.getError().getCode());
        assertEquals("Server Error", exception.getError().getReason());
        assertEquals("Unexpected error deleting Push subscription", exception.getError().getMessage());
        verify(eventGridManager, times(1)).eventSubscriptions();
    }

    @Test
    public void shouldGetPushSubscription() {

        when(eventSubscriptions.getAsync(scope, subscriptionId)).thenReturn(observableGetSubscription);

        pushSubscription.checkIfSubscriptionExists(subscriptionId, topicName);

        verify(eventGridManager, times(1)).eventSubscriptions();
        verify(logger, times(1)).debug(String.format("Push Subscription with id %s fetched successfully", subscriptionId));
    }

    @Test
    public void shouldThrowNullPointerExceptionWhenFetchingPushSubscription() {

        when(eventSubscriptions.getAsync(scope, subscriptionId)).thenReturn(observableErrorNullPointerExceptionGetSubscription);

        AppException exception = assertThrows(AppException.class, () -> {
            pushSubscription.checkIfSubscriptionExists(subscriptionId, topicName);
        });

        assertNotNull(exception);
        assertEquals(404, exception.getError().getCode());
        assertEquals("Not found", exception.getError().getReason());
        assertEquals(String.format("Push Subscription with id %s does not exist.", subscriptionId), exception.getError().getMessage());
        verify(eventGridManager, times(1)).eventSubscriptions();
        verify(logger, times(1)).error(String.format("Push Subscription with id %s does not exist.", subscriptionId));
    }

    @Test
    public void shouldThrow500ExceptionWhenFetchingPushSubscription() {

        when(eventSubscriptions.getAsync(scope, subscriptionId)).thenReturn(observableError500ExceptionGetSubscription);

        AppException exception = assertThrows(AppException.class, () -> {
            pushSubscription.checkIfSubscriptionExists(subscriptionId, topicName);
        });

        assertNotNull(exception);
        assertEquals(500, exception.getError().getCode());
        assertEquals("Server Error", exception.getError().getReason());
        assertEquals("Unexpected error while Fetching Push subscription", exception.getError().getMessage());
        verify(eventGridManager, times(1)).eventSubscriptions();
    }
}
