/*
 * Copyright 2017-2020, Schlumberger
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.opengroup.osdu.register.provider.gcp.subscriber;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.cloud.Timestamp;
import com.google.cloud.datastore.Entity;
import com.google.cloud.datastore.FullEntity;
import com.google.cloud.datastore.Key;
import com.google.cloud.datastore.StringValue;
import org.opengroup.osdu.core.common.provider.interfaces.IKmsClient;
import org.opengroup.osdu.register.subscriber.model.*;
import org.opengroup.osdu.register.utils.Constants;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.io.IOException;
import java.time.Instant;

@Component
class ModelEntityHelper {

    static final String NAME = "name";
    static final String CREATED_ON = "createdOn";
    static final String CREATED_BY = "createdBy";
    static final String DESCRIPTION = "description";
    static final String TOPIC = "topic";
    static final String SUBSCRIPTION = "notificationId";
    static final String PUSH_ENDPOINT = "endpoint";
    static final String SECRET_TYPE = "secretType";
    static final String SECRET_VALUE = "secretValue";

    @Autowired
    private IKmsClient kmsClient;

    Entity getUpdatedEntity(Entity existingEntity, Secret newSecret) throws IOException {
        String secretType = newSecret.getSecretType();
        String encryptedSecret = kmsClient.encryptString(newSecret.toString());
        return Entity.newBuilder(existingEntity)
                .set(SECRET_TYPE, secretType)
                .set(SECRET_VALUE, StringValue.newBuilder(encryptedSecret).setExcludeFromIndexes(true).build())
                .build();
    }

    Subscription convertEntityToSubscription(Entity entity) throws IOException {
        Subscription output = new Subscription();
        output.setId(entity.getKey().getName());
        output.setName(entity.getString(NAME));
        output.setDescription(entity.getString(DESCRIPTION));
        output.setTopic(entity.getString(TOPIC));
        output.setNotificationId(entity.getString(SUBSCRIPTION));
        output.setPushEndpoint(entity.getString(PUSH_ENDPOINT));
        output.setCreatedBy(entity.getString(CREATED_BY));
        output.setCreatedOnEpoch(entity.getTimestamp(CREATED_ON).toSqlTimestamp());

        String decryptedValue = kmsClient.decryptString(entity.getString(SECRET_VALUE));
        if (entity.getString(SECRET_TYPE).equals(Constants.GSASecret)) {
            GsaSecret gsaSecret = new GsaSecret();
            String[] splitSecret = decryptedValue.split("`");
            gsaSecret.setValue(new GsaSecretValue(splitSecret[0], splitSecret[1]));
            output.setSecret(gsaSecret);
        } else {
            HmacSecret hmacSecret = new HmacSecret();
            hmacSecret.setValue(decryptedValue);
            output.setSecret(hmacSecret);
        }
        return output;
    }

    FullEntity convertSubscriptionToEntity(Subscription subscription, Key key) throws IOException {
        subscription.setCreatedOnEpoch(java.sql.Timestamp.from(Instant.now()));
        String encryptedString = kmsClient.encryptString(subscription.getSecret().toString());
        FullEntity.Builder entity = FullEntity.newBuilder(key)
                .set(NAME, subscription.getName())
                .set(CREATED_ON, Timestamp.now())
                .set(CREATED_BY, subscription.getCreatedBy())
                .set(TOPIC, subscription.getTopic())
                .set(SUBSCRIPTION, subscription.getNotificationId())
                .set(PUSH_ENDPOINT, subscription.getPushEndpoint())
                .set(SECRET_TYPE, subscription.getSecret().getSecretType())
                .set(SECRET_VALUE, StringValue.newBuilder(encryptedString).setExcludeFromIndexes(true).build())
                .set(DESCRIPTION, StringValue.newBuilder(subscription.getDescription()).setExcludeFromIndexes(true).build());

        return entity.build();
    }

    Subscription getDeserializedSubscription(String str) throws Exception {
        ObjectMapper objectMapper = new ObjectMapper();
        return objectMapper.readValue(str, Subscription.class);
    }
}
