/*
 * Copyright 2017-2020, Schlumberger
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.opengroup.osdu.register.provider.gcp.ddms.datastore;

import com.google.cloud.Timestamp;
import com.google.cloud.datastore.*;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.opengroup.osdu.core.common.model.tenant.TenantInfo;
import org.opengroup.osdu.register.ddms.model.Ddms;
import org.opengroup.osdu.register.ddms.model.RegisteredInterface;
import org.powermock.core.classloader.annotations.PrepareForTest;
import org.powermock.modules.junit4.PowerMockRunner;

import java.util.Collections;
import java.util.List;
import java.util.UUID;

import static junit.framework.TestCase.assertTrue;
import static org.junit.Assert.assertEquals;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Matchers.anyInt;
import static org.mockito.Mockito.*;
import static org.mockito.MockitoAnnotations.initMocks;

@RunWith(PowerMockRunner.class)
@PrepareForTest({Datastore.class, KeyFactory.class, Key.class})
public class DatastoreDdmsRepositoryTest {
    @Mock
    private DatastoreMultiTenantAccess dataStoreTenants;
    @Mock
    private TenantInfo tenantInfo;
    @Mock
    private Datastore datastore;
    @Mock
    private KeyFactory keyFactory;
    @InjectMocks
    @Spy
    private DatastoreDdmsRepository sut = new DatastoreDdmsRepository();

    private final Timestamp createdOn = Timestamp.now();

    @Before
    public void setup() {
        initMocks(this);
        when(datastore.newKeyFactory()).thenReturn(keyFactory);
        when(dataStoreTenants.get(any())).thenReturn(datastore);
    }

    @Test
    public void should_createGetAndDeleteDms_fromDatastore() {
        Ddms input = createDms();

        Key mockKey = mockKey(1L);
        doReturn(mockKey).when(sut).createkey(any(), any());
        mockTransaction(Collections.emptyList(), false);
        Ddms output = sut.create(input);
        assertEquals(output, input);

        Entity datastoreEntity = createEntity(mockKey);
        when(this.datastore.get(mockKey)).thenReturn(datastoreEntity);
        doReturn(input).when(this.sut).convertEntityToDms(any());
        output = sut.get(input.getId());
        assertEquals(output, input);

        Transaction txn = mockTransaction(Collections.singletonList(mockKey), false);
        when(txn.get(mockKey)).thenReturn(datastoreEntity);
        assertTrue(sut.delete(input.getId()));
    }

    private Ddms createDms() {
        Ddms input = new Ddms();
        RegisteredInterface ri = new RegisteredInterface();
        ri.setEntityType("type");
        ri.setSchema(Collections.singletonMap("1", "2"));

        input.setContactEmail("mock-contactEmail");
        input.setDescription("mock-description");
        input.setId(UUID.randomUUID().toString());
        input.setName("mock-name");
        input.setCreatedDateTimeEpoch(createdOn.toSqlTimestamp());
        input.getInterfaces().add(ri);
        return input;
    }

    private Key mockKey(Long id) {
        Key key = mock(Key.class);
        when(key.getKind()).thenReturn("key-name");
        when(key.getNamespace()).thenReturn("namespace");
        when(key.getId()).thenReturn(id);
        when(key.getProjectId()).thenReturn("mock-project");
        when(keyFactory.setKind(any())).thenReturn(keyFactory);
        when(keyFactory.setNamespace(any())).thenReturn(keyFactory);
        when(keyFactory.newKey(anyInt())).thenReturn(key);
        return key;
    }

    private Transaction mockTransaction(List<Key> keys, Boolean isActive) {
        Transaction txn = mock(Transaction.class);
        when(this.datastore.newTransaction()).thenReturn(txn);
        Transaction.Response rsp = mock(Transaction.Response.class);
        when(txn.commit()).thenReturn(rsp);
        when(txn.isActive()).thenReturn(isActive);
        when(rsp.getGeneratedKeys()).thenReturn(keys.isEmpty() ? Collections.emptyList() : keys);
        return txn;
    }

    private Entity createEntity(Key key) {
        return getEntityBuilder(key)
                .build();
    }

    private Entity.Builder getEntityBuilder(Key key) {
        return Entity.newBuilder(key)
                .set("name", "mock-name")
                .set("description", "mock-description")
                .set("contactEmail", "mock-contactEmail")
                .set("createdOn", "createdOn")
                .set("kind", "DDMS")
                .set("namespace", "DE");
    }
}