/*
 * Copyright 2017-2020, Schlumberger
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.opengroup.osdu.register.provider.gcp.subscriber;

import com.google.cloud.Timestamp;
import com.google.cloud.datastore.Entity;
import com.google.cloud.datastore.FullEntity;
import com.google.cloud.datastore.Key;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.junit.MockitoJUnitRunner;
import org.opengroup.osdu.core.common.provider.interfaces.IKmsClient;
import org.opengroup.osdu.register.subscriber.model.GsaSecret;
import org.opengroup.osdu.register.subscriber.model.GsaSecretValue;
import org.opengroup.osdu.register.subscriber.model.HmacSecret;
import org.opengroup.osdu.register.subscriber.model.Subscription;
import org.opengroup.osdu.register.utils.AppServiceConfig;

import java.io.IOException;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.mockito.ArgumentMatchers.eq;
import static org.opengroup.osdu.register.provider.gcp.subscriber.ModelEntityHelper.*;
import static org.powermock.api.mockito.PowerMockito.when;

@RunWith(MockitoJUnitRunner.class)
public class ModelEntityHelperTest {

    @Mock
    IKmsClient kmsClient;
    @Mock
    AppServiceConfig config;
    @InjectMocks
    ModelEntityHelper sut;

    private static final Key KEY = Key.newBuilder("opendes", "kind", "entity").build();

    @Test
    public void should_update_entity_with_new_secret() throws IOException {
        when(kmsClient.encryptString(eq("newValue"))).thenReturn("newEncryptedSecretValue");
        Entity entity = Entity.newBuilder(KEY)
                .set(NAME, "testEntity")
                .set(SECRET_TYPE, "HMAC")
                .set(SECRET_VALUE, "oldSecret").build();

        Entity updatedEntity = sut.getUpdatedEntity(entity, new HmacSecret("newValue"));
        assertEquals("newEncryptedSecretValue", updatedEntity.getString(SECRET_VALUE));
    }

    @Test
    public void should_convert_entity_with_hmac_secret_to_subscriber() throws IOException {
        Timestamp now = Timestamp.now();
        Entity testEntity = Entity.newBuilder(KEY)
                .set(NAME, "testEntity")
                .set(DESCRIPTION, "Description")
                .set(TOPIC, "topic1")
                .set(SUBSCRIPTION, "notificationId")
                .set(PUSH_ENDPOINT, "endpoint")
                .set(CREATED_BY, "test@opengroup.org")
                .set(CREATED_ON, now)
                .set(SECRET_TYPE, "HMAC")
                .set(SECRET_VALUE, "encrypted")
                .build();
        when(kmsClient.decryptString(eq("encrypted"))).thenReturn("decrypted");

        Subscription subscription = sut.convertEntityToSubscription(testEntity);
        assertEquals("entity", subscription.getId());
        assertEquals("test@opengroup.org", subscription.getCreatedBy());
        assertEquals(now.toSqlTimestamp(), subscription.getCreatedOnEpoch());
        assertEquals("testEntity", subscription.getName());
        assertEquals("Description", subscription.getDescription());
        assertEquals("notificationId", subscription.getNotificationId());
        assertEquals("endpoint", subscription.getPushEndpoint());
        assertEquals("HMAC", subscription.getSecret().getSecretType());
        assertEquals("decrypted", ((HmacSecret) subscription.getSecret()).getValue());
    }

    @Test
    public void should_convert_entity_with_gsa_secret_to_subscriber() throws IOException {
        Timestamp now = Timestamp.now();
        Entity testEntity = Entity.newBuilder(KEY)
                .set(NAME, "testEntity")
                .set(DESCRIPTION, "Description")
                .set(TOPIC, "topic1")
                .set(SUBSCRIPTION, "notificationId")
                .set(PUSH_ENDPOINT, "endpoint")
                .set(CREATED_BY, "test@opengroup.org")
                .set(CREATED_ON, now)
                .set(SECRET_TYPE, "GSA")
                .set(SECRET_VALUE, "encrypted")
                .build();
        when(kmsClient.decryptString(eq("encrypted"))).thenReturn("audience`key");

        Subscription subscription = sut.convertEntityToSubscription(testEntity);
        assertEquals("entity", subscription.getId());
        assertEquals("test@opengroup.org", subscription.getCreatedBy());
        assertEquals(now.toSqlTimestamp(), subscription.getCreatedOnEpoch());
        assertEquals("testEntity", subscription.getName());
        assertEquals("Description", subscription.getDescription());
        assertEquals("notificationId", subscription.getNotificationId());
        assertEquals("endpoint", subscription.getPushEndpoint());
        assertEquals("GSA", subscription.getSecret().getSecretType());
        assertEquals("audience", ((GsaSecret) subscription.getSecret()).getValue().getAudience());
        assertEquals("key", ((GsaSecret) subscription.getSecret()).getValue().getKey());
    }

    @Test
    public void should_convert_subscriber_with_hmac_secret_to_entity() throws IOException {
        Subscription subscription = new Subscription();
        subscription.setName("testEntity");
        subscription.setDescription("Description");
        subscription.setTopic("topic1");
        subscription.setNotificationId("notificationId");
        subscription.setPushEndpoint("endpoint");
        subscription.setCreatedBy("test@opengroup.org");
        HmacSecret hmacSecret = new HmacSecret();
        hmacSecret.setValue("decrypted");
        subscription.setSecret(hmacSecret);
        when(kmsClient.encryptString(eq("decrypted"))).thenReturn("encrypted");

        FullEntity entity = sut.convertSubscriptionToEntity(subscription, KEY);
        assertEquals("testEntity", entity.getString(NAME));
        assertEquals("test@opengroup.org", entity.getString(CREATED_BY));
        assertEquals("Description", entity.getString(DESCRIPTION));
        assertEquals("notificationId", entity.getString(SUBSCRIPTION));
        assertEquals("endpoint", entity.getString(PUSH_ENDPOINT));
        assertEquals("HMAC", entity.getString(SECRET_TYPE));
        assertEquals("encrypted", entity.getString(SECRET_VALUE));
        assertNotNull(entity.getTimestamp(CREATED_ON));
    }

    @Test
    public void should_convert_subscriber_with_gsa_secret_to_entity() throws IOException {
        Subscription subscription = new Subscription();
        subscription.setName("testEntity");
        subscription.setDescription("Description");
        subscription.setTopic("topic1");
        subscription.setNotificationId("notificationId");
        subscription.setPushEndpoint("endpoint");
        subscription.setCreatedBy("test@opengroup.org");
        GsaSecret gsaSecret = new GsaSecret();
        gsaSecret.setValue(new GsaSecretValue("audience", "key"));
        subscription.setSecret(gsaSecret);
        when(kmsClient.encryptString(eq("audience`key"))).thenReturn("encrypted");

        FullEntity entity = sut.convertSubscriptionToEntity(subscription, KEY);
        assertEquals("testEntity", entity.getString(NAME));
        assertEquals("test@opengroup.org", entity.getString(CREATED_BY));
        assertEquals("Description", entity.getString(DESCRIPTION));
        assertEquals("notificationId", entity.getString(SUBSCRIPTION));
        assertEquals("endpoint", entity.getString(PUSH_ENDPOINT));
        assertEquals("GSA", entity.getString(SECRET_TYPE));
        assertEquals("encrypted", entity.getString(SECRET_VALUE));
        assertNotNull(entity.getTimestamp(CREATED_ON));
    }
}
