/* Licensed Materials - Property of IBM              */		
/* (c) Copyright IBM Corp. 2020. All Rights Reserved.*/	

package org.opengroup.osdu.register.provider.ibm.subscriber;

import static com.cloudant.client.api.query.Expression.eq;
import static java.lang.String.format;

import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;

import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.tenant.TenantInfo;
import org.opengroup.osdu.register.provider.ibm.ddms.datastore.DatastoreMultiTenantAccess;
import org.opengroup.osdu.register.subscriber.model.GsaSecret;
import org.opengroup.osdu.register.subscriber.model.GsaSecretValue;
import org.opengroup.osdu.register.subscriber.model.HmacSecret;
import org.opengroup.osdu.register.subscriber.model.Secret;
import org.opengroup.osdu.register.subscriber.model.Subscription;
import org.opengroup.osdu.register.utils.Constants;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;

import com.cloudant.client.api.Database;
import com.cloudant.client.api.query.QueryBuilder;
import com.cloudant.client.api.query.QueryResult;
import com.cloudant.client.org.lightcouch.DocumentConflictException;
import com.cloudant.client.org.lightcouch.NoDocumentException;

import lombok.extern.slf4j.Slf4j;

@Slf4j
@Repository
public class DatastoreAccess implements IDatastoreAccess {

	private static final String SUBSCRIBER_DB_NAME = "subscription";
	/*
	 * @Autowired private PushSubscription pushSubscription;
	 */
	@Autowired
	private DatastoreMultiTenantAccess dataStoreTenants;
	@Autowired
	private TenantInfo tenantInfo;

	@Override
	public Subscription create(Subscription s) {
		s = createSubscriberInDatastore(s);
		try {
			// pushSubscription.create(s, tenantInfo, this.config.getGoogleCloudProject(),
			// config.getServiceAccountIdentity());
		} catch (Exception e) {
			deleteFromDatastore(s.getId(), false);
			throw new AppException(500, "Server Error", "Unexpected error creating subscription", e);
		}
		return s;
	}

	@Override
	public Subscription get(String id) {
		Database db = dataStoreTenants.get(getDbNameWithTenant());

		Subscription subscription = null;
		try {
			SubscriptionDoc subscriptionDoc = db.find(SubscriptionDoc.class, id);
			subscription = convertToSubscription(subscriptionDoc);
		} catch (NoDocumentException e) {
			throw new AppException(e.getStatusCode(), e.getReason(),
					String.format("Subscriber with id %s does not exist.", id));
		} catch (Exception e) {
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		return subscription;

	}

	// TODO: Remove this api when switch notification to use cache
	@Override
	public List<Subscription> query(String notificationId) {
		List<Subscription> output = new ArrayList<>();
		Database db = dataStoreTenants.get(getDbNameWithTenant());
		QueryResult<SubscriptionDoc> results = null;
		try {
			results = db.query(new QueryBuilder(eq("notificationId", notificationId)).build(), SubscriptionDoc.class);

			if (results == null || results.getDocs().isEmpty()) {
				throw new AppException(404, "Not Found",
						format("Sunscriptio not forund for notification id %s ", notificationId));
			}
		} catch (Exception e) {
			log.info("Unexpected error querying subscription");
			e.printStackTrace();
		}
		output = results.getDocs().stream().map(i -> convertToSubscription(i)).collect(Collectors.toList());
		return output;
	}

	@Override
	public boolean delete(String id) {
		return deleteFromDatastore(id, true);
	}

	@Override
	public boolean patch(String id, Secret secret) {
		return updateInDataStore(id, secret);
	}

	@Override
	public List<Subscription> getAll() throws Exception {
		Database db = dataStoreTenants.get(getDbNameWithTenant());
		List<SubscriptionDoc> subscriptionDocList = db.getAllDocsRequestBuilder().includeDocs(true).build()
				.getResponse().getDocsAs(SubscriptionDoc.class);
		List<Subscription> subscriptionsList = subscriptionDocList.stream().map(i -> convertToSubscription(i))
				.collect(Collectors.toList());
		return subscriptionsList;
	}

	private Subscription createSubscriberInDatastore(Subscription s) {
		Database db = dataStoreTenants.get(getDbNameWithTenant());
		SubscriptionDoc subscriptionDoc = new SubscriptionDoc(s);
		try {
			db.save(subscriptionDoc);
		} catch (DocumentConflictException e) {
			throw new AppException(409, "Conflict",
					"A subscriber already exists with the same topic and endpoint combination");
		} catch (Exception e) {

		}
		return convertToSubscription(subscriptionDoc);
	}

	private boolean deleteFromDatastore(String id, boolean deleteSubscription) {
		Database db = dataStoreTenants.get(getDbNameWithTenant());
		boolean output = false;
		if (!db.contains(id)) {
			return output;
		}
		try {
			SubscriptionDoc subscriptionDoc = db.find(SubscriptionDoc.class, id);
			/*
			 * if(deleteSubscription) { // TODO Add flow for deleting the subscription to
			 * the topic //identify first parameter serviceProjectId
			 * //pushSubscription.delete(servicesProjectId,
			 * subscriptionDoc.getNotificationId()); }
			 */
			db.remove(subscriptionDoc);
			output = true;
		} catch (NoDocumentException e) {
			throw new AppException(e.getStatusCode(), e.getReason(), e.getMessage());
		} catch (DocumentConflictException e) {
			throw new AppException(e.getStatusCode(), e.getReason(), e.getMessage());
		}
		return output;
	}

	private boolean updateInDataStore(String id, Secret secret) {
		Database db = dataStoreTenants.get(getDbNameWithTenant());
		boolean output = false;
		try {
			SubscriptionDoc subscriptionDoc = db.find(SubscriptionDoc.class, id);
			subscriptionDoc.setSecretType(secret.getSecretType());
			subscriptionDoc.setSecretValue(secret.toString());
			db.update(subscriptionDoc);
			output = true;
		} catch (NoDocumentException | DocumentConflictException e) {
			throw new AppException(e.getStatusCode(), e.getReason(), e.getMessage());
		} catch (Exception e) {
			log.error("Unexpected Error updating subscription " + e);
		}
		return output;
	}

	/**
	 * @return
	 */
	private String getDbNameWithTenant() {
		return tenantInfo.getName() + "-" + SUBSCRIBER_DB_NAME;
	}

	private Subscription convertToSubscription(SubscriptionDoc doc) {
		String secretValue = doc.getSecretValue();
		Secret secret;
		if (doc.getSecretType().equals(Constants.GSASecret)) {
			GsaSecret gsaSecret = new GsaSecret();
			String[] splitSecret = secretValue.split("`");
			gsaSecret.setValue(new GsaSecretValue(splitSecret[0], splitSecret[1]));
			secret = gsaSecret;
		} else {
			HmacSecret hmacSecret = new HmacSecret();
			hmacSecret.setValue(secretValue);
			secret = hmacSecret;
		}
		return new Subscription(doc.get_id(), doc.getName(), doc.getDescription(), doc.getTopic(),
				doc.getPushEndpoint(), doc.getCreatedBy(), doc.getCreatedOnEpoch(),
				doc.getNotificationId(), secret);
	}

}
