/*
  Copyright 2021 Google LLC
  Copyright 2021 EPAM Systems, Inc

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */

package org.opengroup.osdu.register.provider.reference.action;

import com.google.gson.Gson;
import java.util.List;
import java.util.stream.Collectors;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.http.HttpStatus;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.util.Crc32c;
import org.opengroup.osdu.register.action.model.Action;
import org.opengroup.osdu.register.provider.interfaces.action.IActionRepo;
import org.opengroup.osdu.register.provider.reference.mapper.ActionMapper;
import org.opengroup.osdu.register.provider.reference.model.document.ActionDocument;
import org.opengroup.osdu.register.provider.reference.repository.mongo.ActionDocumentRepository;
import org.springframework.stereotype.Repository;

@Slf4j
@Repository
@RequiredArgsConstructor
public class ReferenceActionRepository implements IActionRepo {

  private static final String CONFLICT_REASON = "Conflict";
  private static final String CONFLICT_MESSAGE = "An action already exists with the same message and endpoint combination.";
  private static final String NOT_FOUND_REASON = "Not found";
  private static final String NOT_FOUND_MESSAGE = "Action with id %s does not exist.";
  private static final String INTERNAL_SERVER_ERROR_REASON = "Internal server error";
  private static final String INTERNAL_SERVER_ERROR_MESSAGE = "Error retrieving action '%s'.";
  private static final String INTERNAL_SERVER_ERROR_MESSAGE_FOR_GET_ALL = "Error retrieving actions.";

  private final ActionDocumentRepository actionDocumentRepository;
  private final ActionMapper actionMapper;

  @Override
  public List<Action> getAllActions() {
    List<ActionDocument> actionDocuments = this.actionDocumentRepository.findAll();

    return actionDocuments.stream().map(document -> {
      Action action = this.actionMapper.toAction(document);
      if (!getActionChecksum(action).equals(document.getCheckSum())) {
        log.error(String.format(INTERNAL_SERVER_ERROR_MESSAGE, action.getName()));
        throw new AppException(HttpStatus.SC_INTERNAL_SERVER_ERROR, INTERNAL_SERVER_ERROR_REASON,
            INTERNAL_SERVER_ERROR_MESSAGE_FOR_GET_ALL);
      }
      return action;
    }).collect(Collectors.toList());
  }

  @Override
  public Action createAction(Action action) {
    if (this.actionDocumentRepository.existsById(action.getId())) {
      throw new AppException(HttpStatus.SC_CONFLICT, CONFLICT_REASON, CONFLICT_MESSAGE);
    }

    ActionDocument actionDocument = this.actionMapper.toDocument(action, getActionChecksum(action));
    ActionDocument saved = this.actionDocumentRepository.save(actionDocument);

    return this.actionMapper.toAction(saved);
  }

  @Override
  public boolean delete(String id) {
    if (this.actionDocumentRepository.existsById(id)) {
      this.actionDocumentRepository.deleteById(id);
      return true;
    }
    return false;
  }

  @Override
  public Action get(String id) {
    ActionDocument actionDocument = this.actionDocumentRepository.findById(id).orElseThrow(
        () -> new AppException(HttpStatus.SC_NOT_FOUND, NOT_FOUND_REASON,
            String.format(NOT_FOUND_MESSAGE, id)));
    Action action = this.actionMapper.toAction(actionDocument);

    if (!getActionChecksum(action).equals(actionDocument.getCheckSum())) {
      throw new AppException(HttpStatus.SC_INTERNAL_SERVER_ERROR, INTERNAL_SERVER_ERROR_REASON,
          String.format(INTERNAL_SERVER_ERROR_MESSAGE, action.getName()));
    }

    return action;
  }

  private String getActionChecksum(Action action) {
    String serializedAction = new Gson().toJson(action);
    return Crc32c.hashToBase64EncodedString(serializedAction);
  }
}
