/*
  Copyright 2021 Google LLC
  Copyright 2021 EPAM Systems, Inc

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */

package org.opengroup.osdu.register.provider.reference.ddms;

import java.util.List;
import java.util.stream.Collectors;
import lombok.RequiredArgsConstructor;
import org.apache.http.HttpStatus;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.register.ddms.model.Ddms;
import org.opengroup.osdu.register.provider.interfaces.ddms.IDdmsRepository;
import org.opengroup.osdu.register.provider.reference.mapper.DdmsMapper;
import org.opengroup.osdu.register.provider.reference.model.document.DdmsDocument;
import org.opengroup.osdu.register.provider.reference.repository.mongo.DdmsDocumentRepository;
import org.springframework.stereotype.Repository;

@Repository
@RequiredArgsConstructor
public class ReferenceDdmsRepository implements IDdmsRepository {

  private static final String CONFLICT_REASON = "Conflict";
  private static final String CONFLICT_MESSAGE = "A DDMS already exists with the same id.";
  private static final String NOT_FOUND_REASON = "Not found";
  private static final String NOT_FOUND_MESSAGE = "Action with id %s does not exist.";

  private final DdmsDocumentRepository ddmsDocumentRepository;
  private final DdmsMapper ddmsMapper;

  @Override
  public Ddms create(Ddms spec) {
    if (this.ddmsDocumentRepository.existsById(spec.getId())) {
      throw new AppException(HttpStatus.SC_CONFLICT, CONFLICT_REASON, CONFLICT_MESSAGE);
    }

    DdmsDocument ddmsDocument = this.ddmsMapper.toDocument(spec);
    DdmsDocument saved = this.ddmsDocumentRepository.save(ddmsDocument);

    return this.ddmsMapper.toDdms(saved);
  }

  @Override
  public Ddms get(String id) {
    DdmsDocument ddmsDocument = this.ddmsDocumentRepository.findById(id).orElseThrow(
        () -> new AppException(HttpStatus.SC_NOT_FOUND, NOT_FOUND_REASON,
            String.format(NOT_FOUND_MESSAGE, id)));
    return this.ddmsMapper.toDdms(ddmsDocument);
  }

  @Override
  public List<Ddms> query(String type) {
    List<DdmsDocument> documentList = this.ddmsDocumentRepository.queryDmsByType(type);
    return documentList.stream().map(this.ddmsMapper::toDdms).collect(Collectors.toList());
  }

  @Override
  public boolean delete(String id) {
    if (this.ddmsDocumentRepository.existsById(id)) {
      this.ddmsDocumentRepository.deleteById(id);
      return true;
    }
    return false;
  }
}
